<?php

namespace Drupal\okta_user_sync\Helper;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\Config;
use Drupal\Core\Config\ImmutableConfig;
use Drupal\Core\Form\FormStateInterface;
use Drupal\okta_user_sync\MoOktaConstants;
use Drupal\user_provisioning\Helpers\moUserProvisioningLogger;
use GuzzleHttp\Client;
use GuzzleHttp\Exception\GuzzleException;
use GuzzleHttp\Exception\RequestException;

use Symfony\Component\HttpFoundation\RedirectResponse;

/**
 * Helper class for okta user sync module.
 */
class MoOktaHelper {

  /**
   * ImmutableConfig property.
   *
   * @var Drupal\Core\Config\ImmutableConfig
   */
  private ImmutableConfig $config;

  /**
   * Config property.
   *
   * @var Drupal\Core\Config\Config
   */
  private Config $configFactory;

  /**
   * Messenger property.
   *
   * @var object
   */
  protected $messenger;

  /**
   * Base URL of the site.
   *
   * @var string
   */
  private $base_url;

  /**
   * URL Path property.
   *
   * @var string
   */
  private string $urlPath;

  /**
   * Logger property.
   *
   * @var Drupal\user_provisioning\Helpers\moUserProvisioningLogger
   */
  private moUserProvisioningLogger $moLogger;

  /**
   * The httpclient property.
   *
   * @var GuzzleHttp\Client
   */
  private Client $http_client;

  /**
   * The userinfo endpoint.
   *
   * @var string
   */
  private $userinfoEndpoint;

  /**
   * Constructs object of MoOktaHelper class.
   */
  public function __construct() {
    global $base_url;
    $this->base_url = $base_url;
    $this->urlPath = $base_url . '/' . \Drupal::service('extension.list.module')->getPath('okta_user_sync') . '/includes';
    $this->config = \Drupal::config('okta_user_sync.settings');
    $this->messenger = \Drupal::messenger();
    $this->configFactory = \Drupal::configFactory()->getEditable('okta_user_sync.settings');
    if ((\Drupal::service('module_handler')->moduleExists('user_provisioning'))) {
      $this->moLogger = new moUserProvisioningLogger();
    }

    $this->http_client = \Drupal::httpClient();

  }

  /**
   * Makes API call to Okta for getting user information.
   *
   * @param string $bearer_token
   *   The bearer token obtained from okta.
   * @param string $api_url
   *   The endpoint to make api call.
   *
   * @return false|string|void
   *   Returns api call response
   */
  public function getUserFromOkta($bearer_token, $api_url) {
    if (!$this->isCurlInstalled()) {
      return Json::encode(
            [
              "statusCode" => 'ERROR',
              "statusMessage" => 'cURL is not enabled on your site. Please enable the cURL module.',
            ]
        );
    }
    $options = ['headers' => ['Authorization' => 'SSWS ' . $bearer_token], 'verify' => FALSE];

    $this->moLogger->addLog('Query url is ' . $api_url, __LINE__, __FUNCTION__, __FILE__);
    $this->moLogger->addFormattedLog($options, __LINE__, __FUNCTION__, __FILE__, 'The header for testing the search request is:');
    try {
      // 409 conflict should be handled through the catch statement
      $response = \Drupal::httpClient()->get($api_url, $options);

    }
    catch (GuzzleException $exception) {

      if ($exception->getCode() == 0) {
        $error_msg = $exception->getMessage();
        $error_code = [
          "%error" => $error_msg,
          "%Description" => "Please Configure the fields correctly.",
        ];
        \Drupal::logger('user_provisioning')->notice('Error:  %error Cause: %Description', $error_code);
        $this->messenger->addError(t($error_msg));
        $this->configFactory
          ->set('mo_okta_attr_list_from_server', '')
          ->save();
        $response = new RedirectResponse($this->base_url . '/admin/config/people/okta_user_sync/overview?tab=drupal-to-okta-configuration');
        $response->send();
        exit();

      }
      else {
        $this->configFactory
          ->set('mo_okta_attr_list_from_server', '')
          ->save();
        $error = [
          '%error' => $exception->getResponse()->getBody()->getContents(),
        ];
        \Drupal::logger('user_provisioning')->notice('Error:  %error', $error);
        foreach ($error as $key => $value) {
          return $value;
        }
      }
    }
    $content = $response->getBody()->getContents();
    return $content;
  }

  /**
   * Creates url for testing.
   *
   * @param string $url
   *   The url.
   * @param string $upn
   *   The upn.
   *
   * @return string
   *   Returns trimmed url.
   */
  public static function creatUrlForTesting($url, $upn) {
    return rtrim($url, '/') . '/' . 'api' . '/' . 'v1' . '/' . 'users' . '/' . $upn;
  }

  /**
   * Fetches user attributes from Okta.
   */
  public function fetchAttributes() {
    $upn = $this->configFactory->get('okta_user_sync_upn');
    $bearer_token = $this->configFactory->get('okta_user_sync_bearer_token');
    $url = $this->configFactory->get('okta_user_sync_base_url');

    $api_url = self::creatUrlForTesting($url, $upn);
    $response = self::getUserFromOkta($bearer_token, $api_url);
    $content = json_decode($response);
    $user_details = $this->moOktaArrayFlattenAttributes($content);

    $user_details_encoded = Json::encode($user_details);

    $this->configFactory
      ->set('mo_okta_attr_list_from_server', $user_details_encoded)
      ->save();

  }

  /**
   * Checks curl is installed or not.
   *
   * @return bool
   *   Returns true if installed if true else false.
   */
  public function isCurlInstalled(): bool {
    return in_array('curl', get_loaded_extensions());
  }

  /**
   * Flattens user attributes received from Okta.
   *
   * @param array $details
   *   The user info array.
   *
   * @return array
   *   Returns flattened array
   */
  public function moOktaArrayFlattenAttributes($details): array {
    $arr = [];
    foreach ($details as $key => $value) {

      if (empty($value)) {
        continue;
      }
      if (!is_object($value)&& !is_array($value)) {
        $arr[$key] = filter_var($value);
      }
      else {
        $this->moOktaArrayFlattenAttributesLvl2($key, $value, $arr);
      }
    }

    return $arr;
  }

  /**
   * Seperates user attributes by | if nested.
   *
   * @param string $index
   *   Key of array.
   * @param array $arr
   *   The array.
   * @param array $haystack
   *   The array of flattened attributes.
   */
  private function moOktaArrayFlattenAttributesLvl2($index, $arr, &$haystack) {
    foreach ($arr as $key => $value) {
      if (empty($value)) {
        continue;
      }

      if (!is_object($value) && !is_array($value)) {

        if (!strpos(strtolower($index), 'error')) {
          $haystack[$index . "|" . $key] = $value;
        }

      }
      else {
        $this->moOktaArrayFlattenAttributesLvl2($index . "|" . $key, $value, $haystack);
      }
    }
  }

  /**
   * Shows customer support icon.
   */
  public function moOktaShowCustomerSupportIcon(array &$form, FormStateInterface $form_state) {
    $form['mo_okta_customer_support_icon'] = [
      '#markup' => t('<a class="use-ajax mo-bottom-corner"  href="CustomerSupportOkta"><img src="' . $this->urlPath . '/mo-customer-support.png" alt="test image"></a>'),
    ];
  }

  /**
   * Sends user feedback.
   */
  public static function oktaUserSyncFeedbackFunction() {
    global $base_url;
    $post         = \Drupal::request()->request->all();
    $skip         = isset($post['skip']) ? 1 : 0;
    $is_cli       = self::drupal_is_cli();
    $config       = \Drupal::config('okta_user_sync.settings');
    $site_mail    = \Drupal::currentUser()->getEmail();
    $modules_info = \Drupal::service('extension.list.module')->getExtensionInfo('okta_user_sync');
    $fromEmail    = '';
    if (isset($post['uninstall_email']) && !empty($post['uninstall_email'])) {
      $fromEmail = $post['uninstall_email'];
    }
    else {
      $fromEmail = $site_mail;
    }

    $customerKey = "16555";
    $apikey      = "fFd2XcvTGDemZvbw1bcUesNJWEqKbbUq";

    $drupal_core         = \DRUPAL::VERSION;
    $mo_version          = $modules_info['version'];
    $php_version         = phpversion();
    $currentTimeInMillis = self::getTimestamp();
    $stringToHash        = $customerKey . $currentTimeInMillis . $apikey;
    $hashValue           = hash("sha512", $stringToHash);

    if ($skip) {
      self::skippedFeedback($mo_version, $fromEmail);
    }
    else {
      $url    = MoOktaConstants::MO_NOTIFY_SEND;
      $reason = '';
      if (isset($post['uninstall_reason'])) {
        $reason = $post['uninstall_reason'];
      }
      elseif ($is_cli) {
        $reason = "Command Line Uninstallation";
      }

      $q_feedback              = $post['uninstall_other_reason'] ?? '-';
      $config                  = \Drupal::configFactory()->getEditable('okta_user_sync.settings');
      $test_manual_create_user = $config->get('okta_user_sync_create_user_manually');
      $create_user             = $config->get('okta_user_sync_create_user');
      $deactivate_user         = $config->get('okta_user_sync_deactivate_user');
      $payment_page_visited    = $config->get('okta_user_sync_payment_page_visited');
      $test_config_result      = $config->get('okta_user_sync_test_config_result');

      if (isset($post['uninstall_other_reason']) && !empty($post['uninstall_other_reason'])) {
        $reason = $reason . ' (' . $q_feedback . ') ';
      }

      $subject = "Feedback for Drupal {$drupal_core} Okta User Sync Free | {$mo_version} | {$php_version} - {$fromEmail}";

      $subject = $is_cli ? "Skipped " . $subject : $subject;

      $query = '[Drupal ' . \DRUPAL::VERSION . ' Okta User Sync | ' . $mo_version . ']: ' . $reason . '<br>';
      $uninstall_data = '<div><b>Payment Page Visited: </b>' . $payment_page_visited . '<br><b>Test Configuration Result: </b>' . $test_config_result . '<br><b>Test Manual Sync: </b>' . $test_manual_create_user . '<br><b>Create User: </b>' . $create_user . '<br><b>Deactivate User: </b>' . $deactivate_user . '</div>';
      $content = '<div >Hello, <br><br>Company :<a href="' . $base_url . '" target="_blank" >' . $base_url . '</a><br><br>Email :<a href="mailto:' . $fromEmail . '" target="_blank">' . $fromEmail . '</a><br><br>Query :' . $query . '<br><br>' . $uninstall_data . '</div>';

      $fields = [
        'customerKey' => $customerKey,
        'sendEmail' => TRUE,
        'email' => [
          'customerKey' => $customerKey,
          'fromEmail' => $fromEmail,
          'fromName' => 'miniOrange',
          'toEmail' => 'drupalsupport@xecurify.com',
          'toName' => 'drupalsupport@xecurify.com',
          'subject' => $subject,
          'content' => $content,
        ],
      ];
      $field_string = json_encode($fields);
      $header = [
        'Content-Type' => 'application/json',
        'Customer-Key' => $customerKey,
        'Timestamp' => $currentTimeInMillis,
        'Authorization' => $hashValue,
      ];

      $response = (new MoOktaHelper)->callService($url, $field_string, $header);
    }
  }

  /**
   * Sends feedback if skipped.
   */
  public static function skippedFeedback($modules_version, $email) {
    global $base_url;
    $url         = MoOktaConstants::MO_NOTIFY_SEND;
    $customerKey = "16555";
    $apikey      = "fFd2XcvTGDemZvbw1bcUesNJWEqKbbUq";

    $currentTimeInMillis = self::getTimestamp();
    $stringToHash        = $customerKey . $currentTimeInMillis . $apikey;
    $hashValue           = hash("sha512", $stringToHash);
    $fromEmail           = 'no-reply@xecurify.com';
    $subject             = 'Regarding Drupal Okta User Sync Module - ' . $modules_version;

    $content =
        '<div>Hello there,
            <br><br>We hope this message finds you well. Recently miniOrange Okta User Sync module has been uninstalled from your site <a href = $base_url target="_blank">(' . $base_url . ')</a>
            <br><br>We understand that uninstalling the module may have been a deliberate decision. If you have any concerns that prompted you to uninstall the module, we would appreciate your feedback as we are constantly working to improve our product.
            <br><br>If you require any assistance with the module or you have encountered any issues while configuring the module that led to the uninstallation, please do not hesitate to reach out to <a href="mailto:' . MoOktaConstants::SUPPORT_EMAIL . '" target="_blank">' . MoOktaConstants::SUPPORT_EMAIL . '</a>, we will be happy to assist you. If you want we can discuss your use-case over a quick call.
            <br><br><b>If you want, we can also provide you with the 7 days free full-featured trial.</b>
            <br><br>We appreciate you trying out our miniOrange Okta User Sync module, and we hope that we can continue to serve you in the future.
            <br><br><p style="font:small/1.5 Arial,Helvetica,sans-serif">Thanks and Regards,
            <br><b>mini<font color="#ff9900">O</font>range Drupal Support</b></p><hr>
            <span align="left" style="padding-top:15px;font-size:10px;line-height:14px;color:#676f71;font-family:Arial,Helvetica,sans-serif">Please do not reply to this email as we are not able to respond to messages sent to this address. For any quries please contact at <a href="mailto:' . MoOktaConstants::SUPPORT_EMAIL . '" >' . MoOktaConstants::SUPPORT_EMAIL . '</a>.
            </span>';
    $fields = [
      'customerKey'    => $customerKey,
      'sendEmail'     => TRUE,
      'email'           => [
        'customerKey'   => $customerKey,
        'fromEmail'     => $fromEmail,
        'fromName'           => 'miniOrange',
        'toEmail'            => $email,
        'toName'             => $email,
//        'bccEmail'      => 'drupalsupport@xecurify.com',
        'subject'           => $subject,
        'content'           => $content,
      ],
    ];

    $field_string = json_encode($fields);

    $header = [
      'Content-Type' => 'application/json',
      'Customer-Key' => $customerKey,
      'Timestamp' => $currentTimeInMillis,
      'Authorization' => $hashValue,
    ];

    $response = (new MoOktaHelper)->callService($url, $field_string, $header);
  }

  /**
   * Checks if module is getting unistalled on cli.
   *
   * @return bool
   *   Retuns true if cli else false
   */
  public static function drupal_is_cli() {
    $server = \Drupal::request()->server;
    $server_software = $server->get('SERVER_SOFTWARE');
    $server_argc = $server->get('argc');

    if (!isset($server_software) && (php_sapi_name() == 'cli' || (is_numeric($server_argc) && $server_argc > 0))) {
      return TRUE;
    }
    else {
      return FALSE;
    }
  }

  /**
   * @param  string $url
   *   The url of make api call.
   * @param  array $fields
   *   The body of api request.
   * @param  bool $header
   *   The header of api request.
   * @return false|string|void
   *   Returns api call response
   */
  public function callService($url, $fields, $header) {
    if (!$this->isCurlInstalled()) {
      return json_encode(
            [
              "statusCode" => 'ERROR',
              "statusMessage" => 'cURL is not enabled on your site. Please enable the cURL module.',
            ]
        );
    }
    $fieldString = is_string($fields) ? $fields : json_encode($fields);

    try {
      $response = \Drupal::httpClient()
        ->post(
            $url, [
              'body' => $fieldString,
              'allow_redirects' => TRUE,
              'http_errors' => FALSE,
              'decode_content' => TRUE,
              'verify' => FALSE,
              'headers' => $header,
            ]
        );
      return $response->getBody()->getContents();
    }
    catch (RequestException $exception) {
      $error = [
        '%apiName' => explode("moas", $url)[1],
        '%error' => $exception->getResponse()->getBody()->getContents(),
      ];
      \Drupal::logger('okta_user_sync')->notice('Error at %apiName of  %error', $error);
    }
  }

  /**
   * This function is written to get the timestamp.
   *
   * @return string
   *   Returns timestamp.
   */
  public static function getTimestamp() {
    $url = MoOktaConstants::GET_TIMESTAMP;
    $response = (new MoOktaHelper)->callService($url, [], []);

    if (empty($content)) {
      $currentTimeInMillis = round(microtime(TRUE) * 1000);
      $currentTimeInMillis = number_format($currentTimeInMillis, 0, '', '');
    }
    return empty($content) ? $currentTimeInMillis : $content;
  }

}
